package oracle.forms.fd;

import java.awt.Component;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;
import oracle.forms.jdapi.*;
import java.io.*;
import java.util.HashMap;
import javax.swing.JOptionPane;


/*************************************************************** 
 *  Forms JAVA API (JDAPI): - <em>JDAPI_LAF</em><br>
 *  This JavaBean allows doing modifications on Forms modules
 *  based on the content of an xml file
 *  to implement the LAF project
 *
 * @author  F. Degrelle
 * creation May-2008
 * @version 1.5.8
 * 
 * added the block's bounding frame remove/replacement
 * added the Current Record Attribute form-level triggers
 * Switch standard LOVs to LAF LOVs
 ***************************************************************/

public class JDAPI_LAF extends Component
{
  // constants
  public final static int OVERRIDE = 0 ;
  public final static int BEFORE   = 1 ;
  public final static int AFTER    = 2 ;
  // variables
  public static String      sModule_name    = "JDAPI_LAF";
  public static String      sModule_version = "1.5.7";
  String      sBackupDir = "" ;   // module backup directory
  String      sTargetDir = "" ;   // module target directory
  String      sModuleName = "";   // module name
  String      sXMLfile = null ;   // XML configuration file name
  String      sBackExt = null ;   // backup extension
  String      sNewExt  = null ;   // new file extension
  String      sLibCase = null ;   // libraries case conversion  
  String      sLogFile = null ;   // log filename
  String      sConnect = "" ;     // connection string  
  String      sFileNew = "" ;     // new filename  
  String      sComment = "" ;
  String      sSleep   = "" ;     // PL/SQL code to wait for a moment
  String      sFileContent = "" ;
  String      sCurrentModule = "" ;
  String      sWhenNewFormInstance = "" ;
  String      sHideButtons = "" ;
  String      sFirstBlock = "" ;
  String      sErrorCode = "^ERROR^" ;
  int         iItemBLevel = 2 ; // none
  int         iTotCanvases = 0 ;
  int         iTotFrames = 0 ;
  float       fConvertRate = 1.0f ;
  HashMap     hCanvases    = new HashMap() ;
  HashMap     hBeans       = new HashMap() ;  
  JdapiModule mod = null ;
  FormModule  fm = null ;  
  List        module_list  = new ArrayList(5) ; 
  List        block_list   = new ArrayList(5) ;   
  boolean     bLog = true ;
  boolean     bLogDetail = false ;
  boolean     bLogOpen = false ;
  boolean     bCompile = false ;
  boolean     bFileOk  = false ;
  boolean     bRemoveBlockFrames = false ;
  boolean     bReplaceBlockFrames = false ;
  boolean     bReplaceLOVs = false ;
  boolean     bLOVsFound = false ;  
  boolean     bUpdCBox = true ;
  boolean     bUpdRButton = true ;
  boolean     bUpdTField = true ;
  boolean     bUpdTArea = true ;
  boolean     bUpdTList = true ;  
  boolean     bShowListComps = false ;
  LOV         TLovs[] = new LOV[20];
  String      sLovs[] = new String[20];
  int         iNbLovs = 0 ;
  PrintWriter fileOut = null;
  File        f = null;
  // path
  String      sPath = null ; // file path
  String      sFile = null ; // file name
  String      sExt = "" ;    // file extension
  int         ie = 0 ;
  // generation date
  Date        today = new Date();
  SimpleDateFormat formater = new SimpleDateFormat("dd/MM/yyyy H:mm:ss");
  String      dateNewFormat = formater.format(today);   

    
  /**--------------------------------------*
   *   public procedure to call.           *
   *   the first argument : sXMLFileName   *
   *   must contain a valid XML filename   *
   *---------------------------------------*/
  public void doProcess(String sXMLFileName)
  {

    try
    {

      if (sXMLFileName == null)
        throw new ErrorException("XML configuration file name must be supplied");      
      
      sXMLfile = sXMLFileName ;
      
      /**------------------------------------------*
       * open and read the xml configuration file  * 
       *-------------------------------------------*/
      log("\nloading XML configuration file: " + sXMLfile);
      sFileContent = readFile(sXMLfile) ;
      if(bFileOk)
      {         
        int j = 0 ;
        String st = "", st2="" ;
        
        /**------------------------------*
         *    configuration  properties  *
         *-------------------------------*/
        st = getXMLTag(sFileContent, "configuration", null) ;
        if(st != null)
        {            

          //
          // log file name
          //
          sLogFile = getXMLTag(sFileContent, "log_file_name", null) ;
          // open the log file
          if(sLogFile != null) 
          { 
            if(! openLogFile(sLogFile)) System.exit(1); 
          }
          else log("!! no log filename given !!");
        
            //Print header
          log("\n*******************************************");
          log("***                                     ***");
          log("***\t"+sModule_name + " Version " + sModule_version+"\t\t***"); 
          log("***\t"+dateNewFormat+"\t\t***"); 
          log("***                                     ***");
          log("*******************************************");
    
          //Print header
          log(" * XML filename       : "+sXMLfile);          
          if(bLogOpen) log(" * logging file       : "+sLogFile);
          
          // ----------------------------
          //  list of modules to process
          // ----------------------------
          st2 = getXMLTag(sFileContent, "module_list", null) ;
          if(st2 == null) throw new ErrorException("<module_list>: no module list given");
          List lst = getXMLTagArray(st2, "name") ;
          if(lst.size()>0)
          {
              module_list  = new ArrayList(5) ; 
              for(j=0; j<lst.size(); j++)
              {
                  String sMod = (String)lst.get(j) ;
                  File fx = new File(sMod); 
                  // directory given ?
                  if(fx.isDirectory())
                  {
                    File files[] = fx.listFiles();
                    for(int k=0; k<files.length; k++)
                    {
                      sFile = files[k].getAbsolutePath();
                      // file extension
                      ie = sFile.indexOf(".") ;
                      sExt = sFile.substring(ie+1);
                      if(sExt.equalsIgnoreCase("FMB"))
                      {
                        // add module to the list
                        log(" > module to process  : "+sFile);
                        module_list.add(sFile);                        
                      }
                    }
                  }
                  else
                  {
                    // add module to the list
                    log(" > module to process  : "+sMod);
                    module_list.add(sMod);
                  }
              }
          }
          else throw new ErrorException("<module_list>: no module list given");
          
          //
          // show component list
          //
          String s1 = getXMLTag(sFileContent, "show_component_list", "false") ;
          log(" * show component list : "+s1);          
          if(s1.equalsIgnoreCase("true")) bShowListComps = true ;
          else bShowListComps = false ;          
          //
          // backup directory
          //
          sBackupDir = getXMLTag(sFileContent, "backup_directory", null) ;
          log(" * backup directory    : "+sBackupDir);

          //
          // target directory
          //
          sTargetDir = getXMLTag(sFileContent, "target_directory", null) ;
          log(" * target directory    : "+sTargetDir);

          //
          // connection string
          //
          sConnect = getXMLTag(sFileContent, "connection_string", null) ;
          log(" * connection string   : "+sConnect);

          //
          // log Detail ?
          //
          st2 = getXMLTag(sFileContent, "log_detail", "false") ;
          if(st2.equalsIgnoreCase("true")) bLogDetail = true ;          
          log(" * detailed log        : "+st2);

          
          //
          // auto compilation
          //
          st2 = getXMLTag(sFileContent, "compile_module", "false") ;
          if(st2.equalsIgnoreCase("true")) bCompile = true ;
          log(" * auto compilation    : "+bCompile);
          
          //
          // backup file extension
          //
          sBackExt = getXMLTag(sFileContent, "backup_ext", "_bak") ;
          log(" * backup extension    : "+sBackExt);
          
          //
          // new file extension
          //
          sNewExt = getXMLTag(sFileContent, "new_ext", "") ;
          log(" * new file extension  : "+sNewExt);
          
          //
          // comment line
          //
          sComment = getXMLTag(sFileContent, "comment", null) ;

          //
          // libraries case conversion
          //
          sLibCase = getXMLTag(sFileContent, "libraries_case", null) ;
          log(" * library conversion  : "+sLibCase);
            
          //
          // sleep instruction
          //
          sSleep = getXMLTag(sFileContent, "sleep_instruction", "dbms_lock.sleep(2/10);") ;
          log(" * sleep instruction   : "+sSleep);
          
          //
          // remove block's bounding frame
          //
          s1 = getXMLTag(sFileContent, "remove_block_frame", "false") ;
          log(" * remove block frame  : "+s1);          
          if(s1.equalsIgnoreCase("true")) bRemoveBlockFrames = true ;
          else bRemoveBlockFrames = false ;

          //
          // replace block's bounding frame
          //
          s1 = getXMLTag(sFileContent, "replace_block_frame", "false") ;
          log(" * replace block frame : "+s1);          
          if(s1.equalsIgnoreCase("true")) bReplaceBlockFrames = true ;
          else bReplaceBlockFrames = false ;
          
          //
          // replace LOVs
          //
          s1 = getXMLTag(sFileContent, "replace_lov", "false") ;
          log(" * replace LOVs        : "+s1);          
          if(s1.equalsIgnoreCase("true")) bReplaceLOVs = true ;
          else bReplaceLOVs = false ;          

          
          //
          // replace Check Boxe
          //
          s1 = getXMLTag(sFileContent, "replace_check_box", "true") ;
          log(" * replace CheckBox    : "+s1);          
          if(s1.equalsIgnoreCase("true")) bUpdCBox = true ;
          else bUpdCBox = false ;                    
          
          //
          // replace Radio Button
          //
          s1 = getXMLTag(sFileContent, "replace_radio_button", "true") ;
          log(" * replace RadioButton : "+s1);          
          if(s1.equalsIgnoreCase("true")) bUpdRButton = true ;
          else bUpdRButton = false ;
          
          //
          // replace TextField
          //
          s1 = getXMLTag(sFileContent, "replace_textfield", "true") ;
          log(" * replace TextField   : "+s1);          
          if(s1.equalsIgnoreCase("true")) bUpdTField = true ;
          else bUpdTField = false ; 
          
          //
          // replace TextArea
          //
          s1 = getXMLTag(sFileContent, "replace_textarea", "true") ;
          log(" * replace TextArea    : "+s1);          
          if(s1.equalsIgnoreCase("true")) bUpdTArea = true ;
          else bUpdTArea = false ;  
          
          //
          // replace TList
          //
          s1 = getXMLTag(sFileContent, "replace_list", "true") ;
          log(" * replace List        : "+s1);          
          if(s1.equalsIgnoreCase("true")) bUpdTList = true ;
          else bUpdTList = false ;                     
          
          //
          // table-block item bevel
          //
          s1 = getXMLTag(sFileContent, "table_block_bevel", "keep") ;
          log(" * table-block bevel   : "+s1);          
          if(s1.equalsIgnoreCase("raised"))       iItemBLevel = 0 ;
          else if(s1.equalsIgnoreCase("lowered")) iItemBLevel = 1 ;
          else if(s1.equalsIgnoreCase("none"))    iItemBLevel = 2 ;
          else if(s1.equalsIgnoreCase("inset"))   iItemBLevel = 3 ;
          else if(s1.equalsIgnoreCase("outset"))  iItemBLevel = 4 ;
          else if(s1.equalsIgnoreCase("plain"))   iItemBLevel = 5 ;
          else if(s1.equalsIgnoreCase("keep"))    iItemBLevel = -1 ;
        }  

        /**------------------------------*
         *   properties for all modules  *
         *-------------------------------*/
        
        st = getXMLTag(sFileContent, "all_modules", null) ;
        if(st != null)
        {            
          // --------------------------------
          //  loop through all given modules
          // --------------------------------
          for(int z=0; z < module_list.size(); z++)
          {
            iTotCanvases = 0 ;
            iTotFrames = 0 ;
            
            // open the module
            String sModule = (String)module_list.get(z);
            sCurrentModule = sModule ;
            if(!openModule(sModule)) continue ;
            // find the Coordinate System
            Coordinate coord = new Coordinate(fm);
            /*
            RealUnit -> 0 = Pixel  
            RealUnit -> 1 = Centimeter  
            RealUnit -> 2 = Inch  
            RealUnit -> 3 = Point  
            RealUnit -> 4 = Decipoint  
            */
            log("   Coordinate System : "+coord.getCoordinateSystem());
            log("   Real unit         : "+coord.getRealUnit());
            switch(coord.getRealUnit())
            {
              case 1: fConvertRate = 37.7949f ; break ;
              case 2: fConvertRate = 96.00249f ; break ;
              case 3: fConvertRate =  1.3333f ; break ;
              case 4: fConvertRate =  .1333f ; break ;
            }
            //
            // make a backup copy
            //
            if(sBackupDir != null)
            {
              log("-- Backuping module to: " + sBackupDir + sFile + sBackExt + "." + sExt);
              fm.save(sBackupDir + sFile + sBackExt + "." + sExt);  
            }
            else
            {
              log("-- Backuping module to: " + sPath + sFile + sBackExt + "." + sExt);
              fm.save(sPath + sFile + sBackExt + "." + sExt);
            }
              
            hCanvases = new HashMap() ;
            hBeans    = new HashMap() ;
            
            // ----------------------
            //  scan for libraries
            // ----------------------
            if(! scanLibraries(sFileContent)) throw new ErrorException(" !! Process aborted !!");          
            
            // --------------------------------------
            //  scan for subclassed objects from olb
            // --------------------------------------
            if(! scanSubclassedObjects(sFileContent)) throw new ErrorException(" !! Process aborted !!");          
                        
            //
            // libraries case conversion
            //
            if(sLibCase != null) setLibraryCaseName(sLibCase);
            
            //
            // All canvases operations
            //
            String sCanvases = getXMLTag(sFileContent, "canvases", null) ;
            if(sCanvases != null)
            {
                // find all the module's canevases
                log2("-- Canvas processing --");
                log3("-- Canvas processing --");
                JdapiIterator cans = fm.getCanvases();
                String sCanvasName = "LAF_BEAN" ;
                //Loop through the canvases
                while(cans.hasNext())  
                { 
                  Canvas cv = (Canvas)cans.next();
                  cCanvas canvas = (cCanvas)hCanvases.get(cv.getName()) ;
                  if(! scanCanvases(sCanvases, cv.getName())) throw new ErrorException(" !! Process aborted !!");
                  iTotCanvases++ ;
                  int iCvType = cv.getCanvasType() ;
                  log2("   Canvas found:"+cv.getName()+" type:"+iCvType);
                  log3("   Canvas found:"+cv.getName()+" type:"+iCvType);
                  // add a java bean to every canvas found
                  if(iCvType == 0 || iCvType == 1 || iCvType == 4) 
                  {
                      try{
                        Block blk = Block.find(fm,"LAF_BLOCK") ;
                        // looking for canvas content, stacked or tab
                        if(iTotCanvases > 1) {
                          sCanvasName = "LAF_BEAN" + iTotCanvases ;
                          if(canvas == null) canvas = new cCanvas(cv.getName(),"LAF_BLOCK." + sCanvasName);

                          hCanvases.put(cv.getName(), canvas);
                          if(iCvType != 4)
                          {
                            // create a new JavaBean component in LAF_BLOCK
                            // for content or stacked canvas
                            Item itm = new Item(blk,sCanvasName);
                            itm.setName(sCanvasName);
                            itm.setComment("Bean Area added for block:"+blk.getName());
                            itm.setItemType(0);
                            itm.setImplementationClass("oracle.forms.fd.DrawLAF");
                            itm.setXPosition(1);
                            itm.setYPosition(1);
                            itm.setWidth(1);
                            itm.setHeight(1);
                            itm.setKeyboardNavigable(false);
                            itm.setMouseNavigate(false);
                            itm.setCanvasName(cv.getName());
                            log2("     Adding Bean Area Name:"+blk.getName()+"."+itm.getName()+"  on canvas:"+cv.getName());
                          }
                          else
                          {
                            JdapiIterator tabs = cv.getTabPages();
                            while(tabs.hasNext())  {
                                // create a new JavaBean component in LAF_BLOCK
                                // for tab canvas
                                TabPage tp = (TabPage)tabs.next() ;
                                Item itm = new Item(blk,"LAF_BEAN"+(iTotCanvases++));
                                itm.setName("LAF_BEAN"+(iTotCanvases++));
                                itm.setItemType(0);
                                itm.setImplementationClass("oracle.forms.fd.DrawLAF");
                                itm.setXPosition(1);
                                itm.setYPosition(1);
                                itm.setWidth(1);
                                itm.setHeight(1);
                                itm.setKeyboardNavigable(false);
                                itm.setMouseNavigate(false);
                                itm.setCanvasName(cv.getName());
                                itm.setTabPageName(tp.getName());
                                log2("     Adding Bean Area Name:"+blk.getName()+"."+itm.getName()+"  on canvas:"+cv.getName());
                            }
                          }  

                        }
                        else {
                            // LAF bean
                            if(canvas == null) canvas = new cCanvas(cv.getName(),"LAF_BLOCK." + sCanvasName);
                            Item itm = Item.find(blk,"LAF_BEAN") ;
                            log2("     Bean Area Name:"+blk.getName()+"."+itm.getName()+"  on canvas:"+cv.getName());
                            if(itm ==null) log2("item LAF_BLOCK.LAF_BEAN not found !!!");
                            else {
                                itm.setCanvasName(cv.getName()); 
                            }
                            // LAF LOV bean
                            if(bReplaceLOVs){
                                Item itm2 = Item.find(blk,"BEAN_LOV") ;
                                log2("     Bean Area Name:"+blk.getName()+"."+itm2.getName()+" on canvas:"+cv.getName());
                                if(itm2 ==null) log2("item LAF_BLOCK.BEAN_LOV not found !!!");
                                else {
                                    itm2.setCanvasName(cv.getName()); 
                                }
                            }
                            hCanvases.put(cv.getName(), canvas);

                        }
                        
                      }
                      catch (Exception bnf) {
                         log2("block LAF_BLOCK not found !!!");
                      }

                    }
                    
                    // find the canvas' graphics objects 
                    JdapiIterator grafs = cv.getGraphicses();
                    String sTitle = "" ;
                    while(grafs.hasNext())  {
                       Graphics g = (Graphics)grafs.next() ;
                       // frames = type = 8 ;
                       if(g.getGraphicsType() == 8)
                       {
                         sTitle = g.getFrameTitle() ;
                         log2("   Graphic found:"+g.getName()+ " ->"+sTitle);                       
                         //printGraphicProperties(g);
                       }
                    }
                }
            } 
            
            //
            //    Find LOVs
            //
            if(bReplaceLOVs) {
                log2("-- LOVs   processing --");
                String sPgmUnit1 = readFile("laf_prepare_lovs_1.txt");
                boolean bContinue = true ;
                if(sPgmUnit1.equalsIgnoreCase("KO")){
                    System.out.println(" !!! Cannot open file: laf_prepare_lovs_1.txt !!!");
                    bContinue = false ;
                }
                String sPgmUnit2 = readFile("laf_prepare_lovs_2.txt");
                if(sPgmUnit2.equalsIgnoreCase("KO")){
                    System.out.println(" !!! Cannot open file: laf_prepare_lovs_2.txt !!!");
                    bContinue = false ;
                }                
                String sPgmUnit3 = readFile("laf_prepare_lovs_3.txt");
                if(sPgmUnit3.equalsIgnoreCase("KO")){
                    System.out.println(" !!! Cannot open file: laf_prepare_lovs_3.txt !!!");
                    bContinue = false ;
                }
                String sTrg1 = readFile("laf_key_listval.txt");
                if(sTrg1.equalsIgnoreCase("KO")){
                    System.out.println(" !!! Cannot open file: laf_key_listval.txt !!!");
                    bContinue = false ;
                }
                String sTrg2 = readFile("laf_when_validate_item.txt");
                if(sTrg2.equalsIgnoreCase("KO")){
                    System.out.println(" !!! Cannot open file: laf_when_validate_item.txt !!!");
                    bContinue = false ;
                } 
                String sTrg3 = readFile("laf_when_new_item_instance.txt");
                if(sTrg3.equalsIgnoreCase("KO")){
                    System.out.println(" !!! Cannot open file: laf_when_new_item_instance.txt !!!");
                    bContinue = false ;
                }                   
                if(bContinue)
                {
                    StringBuffer pu = new StringBuffer();
                    pu.append(sPgmUnit1);
                    JdapiIterator lovs = fm.getLOVs() ;
                    LOVColumnMapping lcm = null ;
                    RecordGroupColumn rgc = null ;
                    String sFirst = null ;
                    String sTmp = null ;
                    while(lovs.hasNext())  {
                       iNbLovs++ ;                       
                       bLOVsFound = true ;
                       sTmp = sPgmUnit2 ;
                       TLovs[iNbLovs] = (LOV)lovs.next(); 
                       RecordGroup rg = TLovs[iNbLovs].getRecordGroupObject() ;
                       sLovs[iNbLovs] = rg.getRecordGroupQuery();   
                       log2("   LOV:"+TLovs[iNbLovs].getName()+"  Query:"+sLovs[iNbLovs]);
                       // colunm type
                       JdapiIterator cols = rg.getRecordGroupColumns();
                       while(cols.hasNext()){
                          rgc = (RecordGroupColumn)cols.next() ;
                          if(sFirst == null) sFirst = rgc.getName();
                          log2("   RG Column type:"+rgc.getName()+" ->"+rgc.getColumnDataType());
                          // 0 CHAR
                          // 1 NUMBER
                          // 2 DATE
                       }
                       // set part 2 of Program Unit : LAF_Prepare_LOVs()
                       String s2 = _replaceAll(TLovs[iNbLovs].getName(),"$","\\$");
                       //sTmp = sTmp.replaceFirst("%1",TLovs[iNbLovs].getName());
                       sTmp = sTmp.replaceFirst("%1",s2);
                       s2 = _replaceAll(TLovs[iNbLovs].getTitle(),"$","\\$");
                       sTmp = sTmp.replaceFirst("%3",s2);
                       //sTmp = sTmp.replaceFirst("%3",TLovs[iNbLovs].getTitle());
                       s2 = _replaceAll(sLovs[iNbLovs],"$","\\$");
                       sTmp = sTmp.replaceFirst("%4",s2.replaceAll("'","''"));
                       sTmp = sTmp.replaceFirst("%5",""+TLovs[iNbLovs].getWidth());
                       sTmp = sTmp.replaceFirst("%6",""+TLovs[iNbLovs].getHeight());
                       sTmp = sTmp.replaceFirst("%7",sFirst);
                       pu.append(sTmp);
                       // column mapping
                       JdapiIterator mapps = TLovs[iNbLovs].getLOVColumnMappings();
                       int i=0;
                       while(mapps.hasNext())  {
                          lcm = (LOVColumnMapping) mapps.next() ;
                          if(!lcm.getReturnItem().equalsIgnoreCase("")){
                              i++;
                              log2("   LOV Mapping:"+lcm.getName()+" ->"+lcm.getReturnItem());                              
                              if(i == 1) pu.append("  --  LOV item mapping  --\n");
                              pu.append("  LR$MappRec("+i+").LOV_COLUMN    := '"+lcm.getName()+"' ;\n");
                              pu.append("  LR$MappRec("+i+").LOV_ITEM1     := '"+lcm.getReturnItem()+"' ;\n");
                          }
                       }
                       pu.append("\n");
                       //  LOV item/column validations 
                       mapps = TLovs[iNbLovs].getLOVColumnMappings();
                       i=0;
                       while(mapps.hasNext())  {
                          lcm = (LOVColumnMapping) mapps.next() ;
                          if(!lcm.getReturnItem().equalsIgnoreCase("")){
                              i++;
                              log2("   LOV Item/validation:"+lcm.getName()+" ->"+lcm.getReturnItem());                              
                              if(i == 1) pu.append("  --  LOV item/column validations  --\n");
                              pu.append("  LR$Valids("+i+").LOV_COLUMN     := '"+lcm.getName()+"' ;\n");
                              pu.append("  LR$Valids("+i+").LOV_ITEM       := '"+lcm.getReturnItem()+"' ;\n");
                          }
                       }
                       pu.append("\n");                       
                       // Items that support this LOV
                       JdapiIterator blks = fm.getBlocks() ;
                       Block blk = null ;
                       Item it = null ;
                       i=0 ;
                       while(blks.hasNext())  {                       
                           blk = (Block)blks.next() ;
                           JdapiIterator its = blk.getItems() ;
                           while(its.hasNext()){
                             it = (Item)its.next() ;
                             if(it.getLovName().equalsIgnoreCase(TLovs[iNbLovs].getName())){
                               i++;
                               if(i == 1)pu.append("  --  LOV items --\n");
                               pu.append("  LR$LovItm("+i+") := '"+blk.getName()+"."+it.getName()+"' ;\n") ;
                               // remove previous standard LOV
                               it.setLovName("");
                             }  
                           }
                           
                       }
                       if(i > 0) pu.append("  LR$LovRec.LOV_ITEMS := LR$LovItm ;  \n");
                       pu.append(sPgmUnit3);
                    }
                    
                   //
                   // create the program units
                   //
                   if(bLOVsFound){
                       pu.append("\n End;\n");
                       ProgramUnit pgm = new ProgramUnit(fm,"LAF_PREPARE_LOVS");
                       pgm.setProgramUnitText(pu.toString());
                   }                    
                    
                }
                if(bLOVsFound) {
                // add/modify form-level triggers
                // key-listval
                try{
                  Trigger trg = Trigger.find(fm,"KEY-LISTVAL");               
                  trg.setTriggerText(sTrg1 + trg.getTriggerText());
                }
                catch (Exception exp) {
                  Trigger trg = new Trigger(fm,"KEY-LISTVAL");   
                  trg.setTriggerText(sTrg1);
                  trg.setExecuteHierarchy(BEFORE);
                }
                // when-new-item-instance
                try{
                  Trigger trg = Trigger.find(fm,"WHEN-NEW-ITEM-INSTANCE");               
                  trg.setTriggerText(sTrg3 + trg.getTriggerText());
                }
                catch (Exception exp) {
                  Trigger trg = new Trigger(fm,"WHEN-NEW-ITEM-INSTANCE");   
                  trg.setTriggerText(sTrg3);
                  trg.setExecuteHierarchy(BEFORE);
                }                
                // when-validate-item
                try{
                  Trigger trg = Trigger.find(fm,"WHEN-VALIDATE-ITEM");               
                  trg.setTriggerText(sTrg2 + trg.getTriggerText());
                }
                catch (Exception exp) {
                  Trigger trg = new Trigger(fm,"WHEN-VALIDATE-ITEM");   
                  trg.setTriggerText(sTrg2);
                  trg.setExecuteHierarchy(BEFORE);
                }
                // pre-form
                if(bReplaceLOVs) {
                    log2("  set PRE/POST-FORMtriggers for LOVs");
                    try{
                      Trigger trg = Trigger.find(fm,"PRE-FORM");               
                      trg.setTriggerText("  --  set the LOVs  --\n  LAF_Prepare_LOVs ;\n" + trg.getTriggerText());
                    }
                    catch (Exception exp) {
                      Trigger trg = new Trigger(fm,"PRE-FORM");   
                      trg.setTriggerText("  --  set the LOVs  --\n  LAF_Prepare_LOVs ;\n");
                      trg.setExecuteHierarchy(BEFORE);
                    } 
                    // post-form
                    String sTrg = "\n  ------------------------------------------------\n"
                                  + "  --  remove LOV descriptions from the package  --\n"
                                  + "  ------------------------------------------------\n"
                                  + "  PKG_DB_LAF_LOV.Remove_LOV( 'ALL_LOVS', :SYSTEM.CURRENT_FORM ) ;\n\n" ;
                    try{
                      Trigger trg = Trigger.find(fm,"POST-FORM");               
                      trg.setTriggerText(sTrg + trg.getTriggerText());
                    }
                    catch (Exception exp) {
                      Trigger trg = new Trigger(fm,"POST-FORM");   
                      trg.setTriggerText(sTrg);
                      trg.setExecuteHierarchy(BEFORE);
                    } 
                }
                }

            }
            
            //
            // All blocks operations
            //
            String sBlocks = getXMLTag(sFileContent, "blocks", null) ;
            if(sBlocks != null)
            {
                log2("-- Block processing --");
                sWhenNewFormInstance = 
                  "------------------------------------------------------\n"
                + "--    Code added for the LAF project integration    --\n"
                + "------------------------------------------------------\n"
                + " "+ sSleep + "\n"
                + " For laf_i In 1 .. 1 Loop" + "\n"                
                + " If PKG_Look_And_Feel.Open_Css(:PARAMETER.PM$CSS_FILENAME) Then\n" 
                + "   If :PARAMETER.PM$GUI is not null Then\n"
                + "      -- read the global GUI properties --\n" ;
                  Set set = hCanvases.keySet() ;
                  Object[] obj = set.toArray() ;
                  if(obj.length > 0) {
                  for(int b=0 ; b < obj.length ; b++)
                  {
                    cCanvas cv = (cCanvas)hCanvases.get(obj[b]) ; 
                    sWhenNewFormInstance += 
                      "      PKG_LOOK_AND_FEEL.Set_GUI_Properties( :PARAMETER.PM$GUI, '" + cv.sBeanName + "' ) ;\n" ;
                  }
                  }
                  else sWhenNewFormInstance += "   null ;\n" ;
                sWhenNewFormInstance += "   End if ;\n" ;
                
                int iBloc = 1 ;
                String sLov = "" ;
                boolean bVisible = false ;
                // find all the module's blocks
                JdapiIterator blks = fm.getBlocks();
                log3("-- item processing --");           
                //Loop through the blocks
                while(blks.hasNext())  
                { 
                  bVisible = false ;
                  block_list = new ArrayList(5) ;
                  Block blk = (Block)blks.next();
                  if(iBloc == 1) sFirstBlock = blk.getName() ;
                  if(! scanBlocks(sBlocks, blk.getName())) throw new ErrorException(" !! Process aborted !!");
                  // update the When-New-Block-Instance trigger content
                  // only for multi-record blocks
                  
                  //
                  // set the LAF Implementation Class
                  //
                  JdapiIterator its = blk.getItems();
                  while(its.hasNext())  
                  {
                     Item it = (Item)its.next();
                     if(it.getCanvasName().length() > 0) {
                       if((it.getItemType() == 2) && bUpdCBox) it.setImplementationClass("oracle.forms.fd.LAF_XP_CBox");
                       if((it.getItemType() == 6) && bUpdTList)
                       {
                          switch(it.getListStyle())
                          {
                            case 0: it.setImplementationClass("oracle.forms.fd.LAF_XP_PopList"); break ;
                            case 1: it.setImplementationClass("oracle.forms.fd.LAF_XP_TList"); break ;
                            case 2: break; // combo box not managed yet
                          }
                       }
                       if(it.getItemType() == 7) it.setImplementationClass("oracle.forms.fd.LAF_XP_Button");  
                       if((it.getItemType() == 8) && bUpdRButton) it.setImplementationClass("oracle.forms.fd.LAF_XP_RadioButton");
                       if(it.getItemType() == 9)
                       {
                         if(it.isMultiLine() && bUpdTArea) it.setImplementationClass("oracle.forms.fd.LAF_XP_TextArea");
                         else if (bUpdTField) it.setImplementationClass("oracle.forms.fd.LAF_XP_TextField");
                       }  
                       if(!it.isVisible() && (it.getImplementationClass().length() > 0)) {
                           // temporarly show hidden items
                            log2("   ! show hidden item: "+blk.getName()+"."+it.getName());
                            sHideButtons += "  Set_Item_Property('" + blk.getName()+"."+it.getName()+
                                "', VISIBLE, PROPERTY_FALSE);\n";
                            it.setVisible(true);                        
                       }
                     }
                  }
                  //if(blk.getRecordsDisplayCount() > 1)
                  //{
                    log2("   processing block: "+blk.getName());
                    //
                    // get every displayed items of the block to find the main associated canvas
                    //
                     its = blk.getItems();
                     //HashMap hCanvases  = new HashMap() ;
                     String sBeanName = "" ;
                     boolean bFound = false ;
                     while(its.hasNext())  
                      { 
                        Item it = (Item)its.next();
                        log3("   item->"+it.getName()+" - "+it.getItemType());           
                        if(it.isEnabled() && it.isVisible()) {
                            String sCV = it.getCanvasName();
                            if(sCV != null && ( ! sCV.equalsIgnoreCase(""))) {
                                bVisible = true ;
                                if(block_list.size()==0)
                                {
                                   cBlock cb = new cBlock(blk.getName(),sCV) ;
                                   cb.iNbre++ ;
                                   block_list.add(cb);
                                }
                                bFound = false ;
                                for(int y=0; y < block_list.size(); y++)
                                {
                                  cBlock cb = (cBlock)block_list.get(y);
                                  if(cb.sName.equalsIgnoreCase(blk.getName()) && cb.sCanvasName.equalsIgnoreCase(sCV))
                                  {
                                     cb.iNbre++ ;
                                     block_list.set(y, cb);
                                     bFound = true ;
                                     break ;
                                  }
                                }
                                if( ! bFound )
                                {
                                   cBlock cb = new cBlock(blk.getName(),sCV) ;
                                   cb.iNbre++ ;
                                   block_list.add(cb);                                    
                                }                                
                                cCanvas cv = (cCanvas)hCanvases.get(sCV);
                                //if(cv == null) cv = new cCanvas(sCV);
                                if(cv != null) {
                                  cv.add();
                                  hCanvases.put(sCV,cv);
                                }
                            }
                            // table-block bevel style
                            if(blk.getRecordsDisplayCount() > 1) {
                              if( iItemBLevel > -1){
                                it.setBevel(iItemBLevel);
                              }
                            }
                        }
                        // LOV replacement ?
                        if(bReplaceLOVs) {
                            sLov = it.getLovName();
                        }
                      }
                      if(! bVisible) continue ;
                      String sWinner = "" ;
                      int iMax = 0 ;
                      for(int y=0; y < block_list.size(); y++)
                      {
                        cBlock cb = (cBlock)block_list.get(y);
                        if(cb.sName.equalsIgnoreCase(blk.getName()))
                        {
                           if(cb.iNbre++ > iMax)
                           {
                             iMax = cb.iNbre ;
                             sWinner = cb.sCanvasName ;
                             cCanvas cv = (cCanvas)hCanvases.get(sWinner);
                             if(cv != null) sBeanName = cv.sBeanName ;
                           }
                        }
                      }

                      log2("   Main canvas for " + blk.getName() + " is : "+ sWinner);  
                      String sScrollBar = "" ; 
                      if(blk.getScrollbarCanvasName() != null &&  
                      (! blk.getScrollbarCanvasName().equalsIgnoreCase(""))) sScrollBar = "true" ;
                      else sScrollBar = "false" ;
                      
                      String sFrameTitle = null ;
                      boolean bLAFframe = false ;
                      int iFrameHeight = 0 ;
                      int iFrameWidth = 0 ;
                      int iFrameXPos = 0 ;
                      int iFrameYPos = 0 ;
                      /*
                       *  replace the block bounding frame ?
                       */
                      if(bReplaceBlockFrames && (!sWinner.equals("")))
                      {
                         Canvas cv = Canvas.find(fm, sWinner);
                         if(cv != null)
                         {
                            JdapiIterator grafs = cv.getGraphicses();
                            //String sTitle = "" ;
                            while(grafs.hasNext())  {
                               Graphics g = (Graphics)grafs.next() ;
                               // frames = type = 8 ;
                               if((g.getGraphicsType() == 8) && (g.getLayoutDataBlockName().equalsIgnoreCase(blk.getName())))
                               {
                                 log2("   Replacing frame:"+g.getName());
                                 if(g.getFrameTitle() != null) sFrameTitle =  g.getFrameTitle() ;
                                 iFrameHeight = (int)(g.getHeight() * fConvertRate);  
                                 iFrameWidth  = (int)(g.getWidth() * fConvertRate);  
                                 iFrameXPos   = (int)(g.getXPosition() * fConvertRate);  
                                 iFrameYPos   = (int)(g.getYPosition() * fConvertRate);  
                                 iTotFrames++ ;
                                 g.destroy();                                 
                                 bLAFframe = true ;
                               }
                            }                         
                         }
                      }                       
                      /*
                       *  removee the block bounding frame ?
                       */
                      else if(bRemoveBlockFrames && (!sWinner.equals("")))
                      {
                         Canvas cv = Canvas.find(fm, sWinner);
                         if(cv != null)
                         {
                            JdapiIterator grafs = cv.getGraphicses();
                            String sTitle = "" ;
                            while(grafs.hasNext())  {
                               Graphics g = (Graphics)grafs.next() ;
                               // frames = type = 8 ;
                               if((g.getGraphicsType() == 8) && (g.getLayoutDataBlockName().equalsIgnoreCase(blk.getName())))
                               {
                                 log2("   Removing frame:"+g.getName());
                                 if(g.getFrameTitle() != null) sFrameTitle = g.getFrameTitle() ;
                                 g.destroy();
                               }
                            }                         
                         }
                      }                    
                        
                      // add/update the When-New-Form-Instance trigger
                      //hCanvases.put(cv.getName(), canvas);
                      String s2 = (String)hBeans.get(sBeanName);
                      if(s2 == null) {
                        hBeans.put(sBeanName,sBeanName);
                        sWhenNewFormInstance += 
                           "   -- paint the canevases --\n"
                         + "   Go_Block('"+ blk.getName() +"');\n"
                         + "   If :PARAMETER.PM$CANVAS is not null Then \n"
                         + "     PKG_LOOK_AND_FEEL.Paint_Canevas(:PARAMETER.PM$CANVAS, '" + sBeanName + "' ) ;\n"
                         + "   End if ;\n" ;
                      }
                      else 
                      {
                        if(blk.getRecordsDisplayCount() > 1) sWhenNewFormInstance += "   Go_Block('"+ blk.getName() +"');\n" ;                        
                      }
                        if(blk.getRecordsDisplayCount() > 1)  {
                        sWhenNewFormInstance += 
                         "   -- paint the block --\n"
                       + "   PKG_LOOK_AND_FEEL.Paint_Block\n"
                       + "   (\n"
                       + "     PC$Block      => '"+ blk.getName() +"'\n"
                       + "    ,PC$BeanName   => '" + sBeanName + "'\n"
                       + "    ,PC$VA_Name    => :PARAMETER.PM$VA\n"
                       + "    ,PC$HeadClass  => :PARAMETER.PM$HEADER\n"
                       + "    ,PC$BodyClass  => :PARAMETER.PM$BODY\n"
                       + "    ,PC$TitleClass => :PARAMETER.PM$TITLE\n" ;
                       if(sFrameTitle == null || bLAFframe )
                          sWhenNewFormInstance +=  "    ,PC$Title      => null\n" ;
                       else
                          sWhenNewFormInstance +=  "    ,PC$Title      => '" + sFrameTitle + "'\n"; 
                       sWhenNewFormInstance += 
                         "    ,PB$ScrollBar  => " + sScrollBar + "\n"
                       + "   ) ;\n" ;
                        }
                       if (bLAFframe)
                       {
                         sWhenNewFormInstance += "  -- Add a LAF Frame to the block -- \n"
                         + "  Set_Custom_Property('" + sBeanName + "',1,'SET_FRAME','Frame" + iTotFrames + ",2," + sFrameTitle + "');\n"
                         + "  Set_Custom_Property('" + sBeanName + "',1,'SET_FRAME_BOUNDS','Frame" + iTotFrames 
                         + "," + (iFrameXPos + 5) 
                         + "," + (iFrameYPos - 15) 
                         + "," + (iFrameWidth + 30) 
                         + "," + (iFrameHeight + 30) + "') ;\n"
                         + "  Set_Custom_Property('" + sBeanName + "',1,'SET_FRAME_TEXT_OPAQUE','Frame" + iTotFrames + ",false');\n"
                         + "  Set_Custom_Property('" + sBeanName + "',1,'SET_FRAME_ROUND_BORDER','Frame" + iTotFrames + ",true');\n" ;
                       }
                    iBloc++ ;
                    
                  } 
                //}   
            } 
  
            //
            // All items operations
            //
            String sItems = getXMLTag(sFileContent, "items", null) ;
            if(sItems != null)
            {
                List l = getXMLTagArray(sItems, "item") ;
                boolean b=true ;
                if(l.size()>0)
                {
                    for(j=0; j<l.size(); j++)
                    {
                        st = (String)l.get(j) ;
                        String sType   = getXMLTag(st, "type", null) ;
                        String sProp   = getXMLTag(st, "property", null) ;
                        String sValue  = getXMLTag(st, "value", null) ;              
                        String sErr    = getXMLTag(st, "on_error", "continue") ;
                        boolean bProcess = false ;
                        if(sType == null) sType = "all" ;                          
                        // find all the module's blocks
                        JdapiIterator blks = fm.getBlocks();
                        //Loop through the blocks
                        while(blks.hasNext())  
                        { 
                          Block blk = (Block)blks.next();
                          // find all the block's items
                          JdapiIterator its = blk.getItems();
                          //Loop through the items
                          while(its.hasNext())  
                          { 
                            bProcess = false ;
                            Item itm = (Item)its.next();
                            int iType = itm.getItemType();
                            // do we have to process this item ?
                            if     (sType.equalsIgnoreCase("all"))                       bProcess = true ;
                            else if(sType.equalsIgnoreCase("BeanArea")    && iType == 0) bProcess = true ;
                            else if(sType.equalsIgnoreCase("Chart")       && iType == 1) bProcess = true ;
                            else if(sType.equalsIgnoreCase("CheckBox")    && iType == 2) bProcess = true ;
                            else if(sType.equalsIgnoreCase("DisplayItem") && iType == 3) bProcess = true ;
                            else if(sType.equalsIgnoreCase("Tree")        && iType == 4) bProcess = true ;
                            else if(sType.equalsIgnoreCase("Image")       && iType == 5) bProcess = true ;
                            else if(sType.equalsIgnoreCase("List")        && iType == 6) bProcess = true ;
                            else if(sType.equalsIgnoreCase("PushButton")  && iType == 7) bProcess = true ;
                            else if(sType.equalsIgnoreCase("RadioButton") && iType == 8) bProcess = true ;
                            else if(sType.equalsIgnoreCase("TextItem")    && iType == 9) bProcess = true ;
                            log3("item->"+itm.getName()+" - "+itm.getItemType()+ "  type="+sType+"  process:"+bProcess);                
                            if(bProcess)
                            {
                              if(! doAction("ITEM", "UPDATE", blk.getName(), itm.getName(), sProp, sValue, sErr)) throw new ErrorException(" !! Process aborted !!");                            
                            }
                          }                   
                        }   
                    }
                }            
            } 

            //
            // All alerts operations
            //
            String sAlerts = getXMLTag(sFileContent, "alerts", null) ;
            if(sAlerts != null)
            {
                // find all the module's alerts
                JdapiIterator alts = fm.getAlerts();
                //Loop through the alerts
                while(alts.hasNext())  
                { 
                  Alert alt = (Alert)alts.next();
                  if(! scanAlerts(sAlerts, alt.getName())) throw new ErrorException(" !! Process aborted !!");
                }   
            }


            //
            // parameters operations
            //
            String sParams = getXMLTag(sFileContent, "single_parameters", null) ;
            if(sParams != null)
            {
                if(! scanParameters(sParams, null)) throw new ErrorException(" !! Process aborted !!");          
            }
                          
            //
            // All windows operations
            //
            String sWindows = getXMLTag(sFileContent, "windows", null) ;
            if(sWindows != null)
            {
                // find all the module's windows
                JdapiIterator wins = fm.getWindows();
                //Loop through the windows
                while(wins.hasNext())  
                { 
                  Window win = (Window)wins.next();
                  if(! scanWindows(sWindows, win.getName())) throw new ErrorException(" !! Process aborted !!");
                }   
            }

            //
            // create or update the When-New-Form-Instance trigger
            //
            sWhenNewFormInstance += "  Go_Block('" + sFirstBlock + "');\n" 
            + "End if;\n" 
            + "End Loop;\n"
            + sHideButtons
            + "------------------------------------------------------\n" 
            + "--  end code added for the LAF project integration  --\n" 
            + "------------------------------------------------------\n" ;
            try{
               Trigger trg = Trigger.find(fm,"WHEN-NEW-FORM-INSTANCE");               
               trg.setTriggerText(sWhenNewFormInstance + trg.getTriggerText());
            }
            catch (Exception exp) {
               Trigger trg = new Trigger(fm,"WHEN-NEW-FORM-INSTANCE");   
               trg.setTriggerText(sWhenNewFormInstance);
            }

            //
            // create or update the When-New-Record-Instance form level trigger
            //
            String s = 
              "---------------------------------------\n"
            + "--  LAF project : colorize each row  --\n"
            + "--                                     \n"
            + "  PKG_LOOK_AND_FEEL.Fill_Table ;\n"
            + "--                                     \n"
            + "--  end LAF project  added code      --\n" 
            + "---------------------------------------\n" ;
            try{
               Trigger trg = Trigger.find(fm,"WHEN-NEW-RECORD-INSTANCE");               
               trg.setTriggerText(s + trg.getTriggerText());
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }
            catch (Exception exp) {
               Trigger trg = new Trigger(fm,"WHEN-NEW-RECORD-INSTANCE");   
               trg.setTriggerText(s);
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }

            //
            // create or update the Post-Query form level trigger
            //
             s="------------------------------------------------------\n"
             + "--  LAF project : colorize each row                 --\n"
             + "--                                                  --\n"
             + "  Default_Value('TRUE','GLOBAL.LAF_QUERY_TIME');     \n" 
             + "  :GLOBAL.LAF_QUERY_TIME := 'TRUE';                  \n" 
             + "  PKG_LOOK_AND_FEEL.Fill_Table;                       \n" 
             + "  :GLOBAL.LAF_QUERY_TIME := 'FALSE';                  \n" 
             + "--                                                  --\n"
             + "--  end LAF project  added code                     --\n"
             + "------------------------------------------------------\n" ;
            try{
               Trigger trg = Trigger.find(fm,"POST-QUERY");               
               trg.setTriggerText(s + trg.getTriggerText());
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }
            catch (Exception exp) {
               Trigger trg = new Trigger(fm,"POST-QUERY");   
               trg.setTriggerText(s);
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }
            
            //
            // create or update the Post-Record form level trigger
            //
            s="------------------------------------------------------\n"
            + "--  LAF project : colorize each row                 --\n"
            + "--                                                  --\n"
            + "  PKG_LOOK_AND_FEEL.Fill_Table(PB$Start => FALSE) ; --\n"
            + "--                                                  --\n"
            + "--  end LAF project  added code                     --\n"
            + "------------------------------------------------------\n" ;
            try{
               Trigger trg = Trigger.find(fm,"POST-RECORD");               
               trg.setTriggerText(s + trg.getTriggerText());
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }
            catch (Exception exp) {
               Trigger trg = new Trigger(fm,"POST-RECORD");   
               trg.setTriggerText(s);
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }

            //
            // create or update the Key-ExeQry form level trigger
            //
            s="------------------------------------------------------\n"
            + "--  LAF project : colorize each row                 --\n"
            + "--                                                  --\n"
            + "  PKG_LOOK_AND_FEEL.laf_execute_query ;             --\n"
            + "--                                                  --\n"
            + "--  end LAF project  added code                     --\n"
            + "------------------------------------------------------\n" ;
            try{
               Trigger trg = Trigger.find(fm,"KEY-EXEQRY");               
               trg.setTriggerText(s + trg.getTriggerText());
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }
            catch (Exception exp) {
               Trigger trg = new Trigger(fm,"KEY-EXEQRY");   
               trg.setTriggerText(s);
               // set Execution Hierarchy to "before"
               trg.setExecuteHierarchy(BEFORE);
            }

            // -------------------
            //  scan for triggers
            // -------------------
            if(! scanTriggers(sFileContent)) throw new ErrorException(" !! Process aborted !!");
            
            // ------------------------
            //  scan for program units
            // ------------------------
            if(! scanProgramUnits(sFileContent)) throw new ErrorException(" !! Process aborted !!");            
            
          fm.setName(sFile + sNewExt);
          if(sComment != null) fm.setComment(sComment.replaceFirst(":SYSDATE",dateNewFormat) + fm.getComment());
          if(sTargetDir != null) sFileNew = sTargetDir + sFile + sNewExt + "." + sExt ;
          else sFileNew = sPath + sFile + sNewExt + "." + sExt ;
          
          log("-- Saving new module to: " + sFileNew);
          fm.save(sFileNew);
          
          // --------------
          // compilation ?
          // --------------
          if(bCompile && (!sConnect.equalsIgnoreCase("") && sConnect != null))
           {boolean b = compileModule(sFileNew);}
           
        }

        
         /**--------------------------------*
          *  properties for single modules  *
          *---------------------------------*/
         String singles = getXMLTag(sFileContent, "single_modules", null) ;
         if(singles != null)
         {
           List l = getXMLTagArray(singles, "module") ;
           boolean b=true ;
           if(l.size()>0)
           {
              for(j=0; j<l.size(); j++)
              {
                 String sModuleContent = (String)l.get(j) ;            
                 String sModuleName = getXMLTag(sModuleContent, "name", null) ;
                 String sCompile = getXMLTag(sModuleContent, "compile", "false") ;
                 if(sModuleName != null)
                 {
                   sCurrentModule = sModuleName ;
                   if(!openModule(sModuleName)) continue ;
                       
                   //
                   // make a backup copy
                   //
                   if(sBackupDir != null)
                   {
                     log("-- Backuping module to: " + sBackupDir + sFile + sBackExt + "." + sExt);
                     fm.save(sBackupDir + sFile + sBackExt + "." + sExt);  
                   }
                   else
                   {
                     log("-- Backuping module to: " + sPath + sFile + sBackExt + "." + sExt);
                     fm.save(sPath + sFile + sBackExt + "." + sExt);
                   }
                     
                   // -------------------
                   //  scan for triggers
                   // -------------------
                   if(! scanTriggers(sModuleContent)) throw new ErrorException(" !! Process aborted !!");
         
                   // ------------------------
                   //  scan for program units
                   // ------------------------
                   if(! scanProgramUnits(sModuleContent)) throw new ErrorException(" !! Process aborted !!");
         
                   // ----------------------
                   //  scan for libraries
                   // ----------------------
                   if(! scanLibraries(sModuleContent)) throw new ErrorException(" !! Process aborted !!");          
         
                   // --------------------------------------
                   //  scan for subclassed objects from olb
                   // --------------------------------------
                   if(! scanSubclassedObjects(sModuleContent)) throw new ErrorException(" !! Process aborted !!");          
         
                   // ---------------------
                   //  scan for canevases
                   // ---------------------
                   if(! scanCanvases(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");          
                     
                   // ---------------------
                   //  scan for windows
                   // ---------------------
                   if(! scanWindows(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");
                     
                   // ---------------------
                   //  scan for alerts
                   // ---------------------
                   if(! scanAlerts(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");
                   
                   fm.setName(sFile + sNewExt);
                   if(sTargetDir != null) sFileNew = sTargetDir + sFile + sNewExt + "." + sExt ;
                   else sFileNew = sPath + sFile + sNewExt + "." + sExt ;
                   log("-- Saving new module to: " + sFileNew);
                   fm.save(sFileNew);
                   // --------------
                   // compilation ?
                   // --------------
                   if(sCompile.equalsIgnoreCase("true") && (!sConnect.equalsIgnoreCase("") && sConnect != null))
                     b = compileModule(sFileNew);
                 
                 }
           
              }
           }   
               
         }

        
        // --------------
        // compilation ?
        // --------------
        if(bCompile && (!sConnect.equalsIgnoreCase("") && sConnect != null))
           {boolean b = compileModule(sFileNew);}
  
      }
            
      
      //Frees the underlying native object. 
      if(fm != null) fm.destroy();  
    
    } // for(int z=0; z < module_list.size(); z++)
   } // if(st != null)    

    catch (ErrorException ee)
     {
       printError(ee.toString());
     }
     
    catch (JdapiException jex)
    {
      printError(jex.toString());
    }
   
    // Clean up
    Jdapi.shutdown();
    closeLogFile();
  }
  
  /***************************************************************************************/
  /**-------------------*
   *     M  A  I  N     *
   *--------------------*/
  // to test from JDev for instance
  public static void main(String[] args)
  {
    String sXMLFile = "c:/jdapi_laf.xml" ;
    //System.out.println("JDAPI_LAF launched by main() procedure:");
    if(args.length > 0) sXMLFile = args[0];
    JDAPI_LAF jd = new JDAPI_LAF();
    jd.doProcess(sXMLFile);
  }
  /***************************************************************************************/
  
    /**---------------------------------------------*
     *  public procedure to switch logging on/off   *
     *----------------------------------------------*/
    public void setLog(boolean b)
    {
      bLog = b ;
    }
  
    /**---------------------------------------------*
     *  public procedure to switch logging on/off   *
     *----------------------------------------------*/
    public String getLogContent()
    {
      log("getLogContent()"+sLogFile);
      return readFile(sLogFile) ;
    }
  
    /**---------------------------------------------*
     *  public procedure to set the log file name   *
     *----------------------------------------------*/
    public void setLogFilename(String sFilename)
    {
      sLogFile = sFilename ;
    }
    
    /**-----------------------*
     *   manage all actions   *
     *------------------------*/
     boolean doAction(String s1, String s2, String s3, String s4, String s5, String s6, String s7)
    {
      if(s1.equalsIgnoreCase("TRIGGER"))
      {
         if(s2.equalsIgnoreCase("ADD"))              return addTrigger(s3, s5, s6, s7);
         else if(s2.equalsIgnoreCase("REPLACE"))     return updTrigger(s2,s3, s4, s5, s7, "");
         else if(s2.equalsIgnoreCase("HIERARCHY"))   return updTrigger(s2,s3, "", "", s7, s5);
         else if(s2.equalsIgnoreCase("DELETE_CODE")) return updTrigger(s2,s3, "", s5, s7, "");
         else if(s2.equalsIgnoreCase("REMOVE"))      return rmTrigger(s3, s7);        
      }
      else if(s1.equalsIgnoreCase("PROCEDURE")) 
      {
         if(s2.equalsIgnoreCase("ADD"))              return addProc(s3, s5, s6, s7);
         else if(s2.equalsIgnoreCase("REPLACE"))     return updProc(s2, s3, s4, s5, s7);
         else if(s2.equalsIgnoreCase("DELETE_CODE")) return updProc(s2, s3, "", s5, s7);
      }
      else if(s1.equalsIgnoreCase("LIBRARY")) 
      {
         if(s2.equalsIgnoreCase("ATTACH"))           return addLib(s3, s7);
         else if(s2.equalsIgnoreCase("DETACH"))      return rmLib(s3, s7);
      }
      else if(s1.equalsIgnoreCase("OLB")) 
      {
         return addOlb(s3, s4, s5, s6, s7);
      }
      else if(s1.equalsIgnoreCase("CANVAS")) 
      {
         if(s2.equalsIgnoreCase("UPDATE"))   return updCanvas(s3, s4, s5, s6);
      }      
      else if(s1.equalsIgnoreCase("WINDOW")) 
      {
         if(s2.equalsIgnoreCase("UPDATE"))   return updWindow(s3, s4, s5, s6);
      }
      else if(s1.equalsIgnoreCase("ALERT")) 
      {
         if(s2.equalsIgnoreCase("UPDATE"))   return updAlert(s3, s4, s5, s6);
      }
      else if(s1.equalsIgnoreCase("PARAMETER")) 
      {
         return updParameter(s2, s3, s5, s6);
      }      
      else if(s1.equalsIgnoreCase("BLOCK")) 
      {
         //if(s2.equalsIgnoreCase("ADD"))             return addProc(s3, s5, s6, s7);
         if(s2.equalsIgnoreCase("UPDATE"))     return updBlock(s3, s4, s5, s6);
        // else if(s2.equalsIgnoreCase("REMOVE"))     return updProc(s2, s3, "", s5, s7);
      }      
      else if(s1.equalsIgnoreCase("ITEM")) 
      {
         if(s2.equalsIgnoreCase("ADD"))             return addProc(s3, s5, s6, s7);
         else if(s2.equalsIgnoreCase("UPDATE"))     return updItem(s3, s4, s5, s6, s7);
         else if(s2.equalsIgnoreCase("REMOVE"))     return updProc(s2, s3, "", s5, s7);
      }      
      return true ;
    }
    
    /**----------------*
     *  add a trigger  *
     *-----------------*/
    boolean addTrigger(String sName, String sText, String sPos, String sErr)
    {
         try
          {
            Trigger trg = Trigger.find(fm,sName);
            String tri_txt = trg.getTriggerText(); 
            if(sPos.equalsIgnoreCase("start")) 
               tri_txt = sText + "\n" + tri_txt ;
            else
               tri_txt = tri_txt + "\n" +  sText ;
            trg.setTriggerText (tri_txt);      
          }
          catch (Exception e) 
          {            
            // Create the trigger 
            try 
            { 
              Trigger trg = new Trigger(fm, sName ); 
              trg.setTriggerText(sText); 
            } 
            catch( Exception exc) 
            { 
                  System.setErr(System.out);
                  printError("addTrigger() Unable to create trigger: "+sName+". Error details: " + exc.toString());
                  //exc.printStackTrace(); 
                  if(sErr.equalsIgnoreCase("stop"))  return false ;
            }  
         }
         return true ;
    }
    

    /**-------------------*
     *  update a trigger  *
     *--------------------*/
    boolean updTrigger(String sAction, String sName, String sReplace, String sText,String sErr, String sHierarchy)
    {
         try
          {
            Trigger trg = Trigger.find(fm,sName);
            String txt = "" ;
            if((sAction.equalsIgnoreCase("REPLACE")) &&  ((sReplace != null) && (sText != null)))
            {
              //txt = trg.getTriggerText().replaceAll(sReplace, sText); 
              txt = _replaceAll(trg.getTriggerText(), sReplace, sText); 
              trg.setTriggerText (txt);      
            }
            if((sAction.equalsIgnoreCase("DELETE_CODE")) &&  (sText != null))
            {
              //log("updTrigger() deleting code:"+sText);
              //txt = trg.getTriggerText().replaceAll(sText, " "); 
              txt = _replaceAll(trg.getTriggerText(),sText, " "); 
              trg.setTriggerText (txt);      
            }            
            if(sHierarchy != null)
            {
              if(sHierarchy.equalsIgnoreCase("BEFORE"))        trg.setExecuteHierarchy(BEFORE);
              else if(sHierarchy.equalsIgnoreCase("AFTER"))    trg.setExecuteHierarchy(AFTER);
              else if(sHierarchy.equalsIgnoreCase("OVERRIDE")) trg.setExecuteHierarchy(OVERRIDE);
            }
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updTrigger() Unable to find trigger: "+sName+". Error details: " + e.toString());
              //e.printStackTrace(); 
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }


    /**---------------------*
     *  update a parameter  *
     *----------------------*/
    boolean updParameter(String sAction, String sName, String sText, String sErr)
    {

         try
          {
            ModuleParameter prm = ModuleParameter.find(fm,sName);
            if(sAction.equalsIgnoreCase("UPDATE"))
            {
              if(sText != null) prm.setParameterInitializeValue(sText);
              else prm.setParameterInitializeValue("");
            }
            if(sAction.equalsIgnoreCase("REMOVE"))
            {
              prm.destroy();
            }            
          }
          catch (Exception e) 
          {            
              if(sAction.equalsIgnoreCase("ADD")) {
                  ModuleParameter prm = new ModuleParameter(fm,sName);
                  prm.setParameterInitializeValue(sText);
              }
              else
              {
                 System.setErr(System.out);
                 printError("updParameter() Unable to find parameter: "+sName+". Error details: " + e.toString());
                 if(sErr.equalsIgnoreCase("stop"))  return false ;
              }
         }
         return true ;
    }

    /**-------------------*
     *  remove a trigger  *
     *--------------------*/
    boolean rmTrigger(String sName, String sErr)
    {
         try
          {
            Trigger trg = Trigger.find(fm,sName);
            trg.destroy();
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("rmTrigger() Unable to find trigger: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }    
    
    /**---------------------*
     *  add a program unit  *
     *----------------------*/
    boolean addProc(String sName, String sText, String sPos, String sErr)
    {
         try
          {
            ProgramUnit pu = ProgramUnit.find(fm,sName);
            String txt = pu.getProgramUnitText(); 
            if(sPos.equalsIgnoreCase("start")) 
               txt = sText + "\n" + txt ;
            else
               txt = txt + "\n" +  sText ;
            pu.setProgramUnitText(txt);
          }
          catch (Exception e) 
          {            
            // Create the program unit
            try 
            { 
              ProgramUnit pu = new ProgramUnit(fm,sName);
              pu.setProgramUnitText(sText);
            } 
            catch( Exception exc) 
            { 
                  System.setErr(System.out);
                  printError("addProc() Unable to create program unit: "+sName+". Error details: " + exc.toString());
                  //exc.printStackTrace(); 
                  if(sErr.equalsIgnoreCase("stop"))  return false ;
            }  
         }
         return true ;
    }
    

    /**------------------------*
     *  update a program unit  *
     *-------------------------*/
    boolean updProc(String sAction, String sName, String sReplace, String sText, String sErr)
    {
         try
          {
            ProgramUnit pu = ProgramUnit.find(fm,sName);
            String txt = pu.getProgramUnitText(); 
            if( sAction.equalsIgnoreCase("REPLACE") && ((sReplace != null) && (sText != null)))
            {
                txt = _replaceAll(pu.getProgramUnitText(),sReplace, sText); 
                pu.setProgramUnitText(txt);
            }
            if( sAction.equalsIgnoreCase("DELETE_CODE") && (sText != null) )
            {
                txt = _replaceAll(pu.getProgramUnitText(),sText, " "); 
                pu.setProgramUnitText(txt);
            }
          }
          catch (Exception exc) 
          {            
              System.setErr(System.out);
              printError("updProc() Unable to find program unit: "+sName+". Error details: " + exc.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }

    /**-------------------*
     *  attach a library  *
     *--------------------*/
    boolean addLib(String sName, String sErr)
    {
         try
          {
            AttachedLibrary lib1 = new AttachedLibrary(fm,sName); 
            log("-- library attached: "+sName);
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("addLib() Unable to attach library: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }    

    /**--------------------*
     *  detach a library   *
     *---------------------*/
    boolean rmLib(String sName, String sErr)
    {
         try
          {
              AttachedLibrary lib = AttachedLibrary.find(fm,sName);
              lib.destroy();
              log("-- library removed: "+sName);
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("rmLib() Unable to dettach library: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    } 


    /**----------------------------------*
     *  add an Object library component  *
     *-----------------------------------*/
    boolean addOlb(String sName, String sTab, String sObject, String sLabel, String sErr)
    {
         int i = 1 ;
         if(sLabel.equalsIgnoreCase("") || sLabel== null)
         {
           printError("addOlb() ["+sName+"/"+sTab+"]: target name cannot be null");
           if(sErr.equalsIgnoreCase("stop"))  return false ;
           else return true ;
         }  
         try
          {
            int iPos = 0;//Integer.parseInt(sObject) ;
            JdapiObject obj1 = null;
            // subclass from object library 
            ObjectLibrary obl1=ObjectLibrary.open(sName); 
            i=2;
            ObjectLibraryTab tab1 =ObjectLibraryTab.find(obl1,sTab); 
            i=3;
            JdapiIterator it = tab1.getOwnedObjects();
            oracle.forms.jdapi.ObjectGroup og = null ;
            while(it.hasNext()) {
                iPos++;
                Object ob = it.next() ;
                if(ob instanceof oracle.forms.jdapi.ObjectGroup ) {
                    og = (oracle.forms.jdapi.ObjectGroup)ob ;
                    if(og.getName().equalsIgnoreCase(sLabel)) {
                      obj1 = (JdapiObject)og ;
                      break ;
                    }  
                }
            }
            
            obj1.createSubclassedChild( fm,sLabel); 
            log("-- addOlb() OLB component subclassed from:\n\t"+sName+"/"+sTab+"/"+sLabel);            
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              String sLevel = "" ;
              switch(i)
              {
                case 1: sLevel = "[Object library not found:"+sName+"]"; break;
                case 2: sLevel = "[Tab not found:"+sTab+"]"; break;
                case 3: sLevel = "[Component not found:"+sObject+"]"; break;
                case 4: sLevel = "[Subclassing]"; break;
              }
              printError("addOlb() "+sLevel+": "+sName+"-"+sTab+"-"+sObject+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }    
    
    /**------------------*
     *   update canvas   *
     *-------------------*/
     boolean updCanvas(String sName, String sProperty, String sValue, String sErr) {
         //log("updCanvas "+sName+"-"+sProperty+":"+sValue);
         try
          {
            Canvas cv = Canvas.find(fm, sName);
            CanvasObject canvas = new CanvasObject();
            // set the canvas property
            boolean b = canvas.setProperty(cv,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updCanvas() Unable to find canvas: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
     }
    
    
    /**------------------*
     *   update alert   *
     *-------------------*/
     boolean updAlert(String sName, String sProperty, String sValue, String sErr) {
         //log("updAlert "+sName+"-"+sProperty+":"+sValue);
         try
          {
            Alert alt = Alert.find(fm, sName);
            AlertObject alert = new AlertObject();
            // set the alert property
            boolean b = alert.setProperty(alt,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updAlert() Unable to find alert: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
     }
    
    /**------------------*
     *   update window   *
     *-------------------*/
     boolean updWindow(String sName, String sProperty, String sValue, String sErr) {
         //log("updWindow "+sName+"-"+sProperty+":"+sValue);
         try
          {
            Window win = Window.find(fm, sName);
            WindowObject window = new WindowObject();
            // set the window property
            boolean b = window.setProperty(win,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updWindow() Unable to find window: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
     }


    /**-------------------*
     *   update a block   *
     *--------------------*/
    boolean updBlock(String sBlock, String sProperty, String sValue, String sErr)
    {
         try
          {
            Block blk = Block.find(fm, sBlock);
            BlockObject block = new BlockObject();
            // set the block property
            boolean b = block.setProperty(blk,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updBlock() Unable to find block: "+sBlock+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }


    /**-------------------*
     *   update an item   *
     *--------------------*/
    boolean updItem(String sBlock, String sName, String sProperty, String sValue, String sErr)
    {
         try
          {
            Block blk = Block.find(fm, sBlock);
            Item itm = Item.find(blk,sName);
            ItemObject io = new ItemObject();
            // set the item property
            boolean b = io.setProperty(itm,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updItem() Unable to find item: "+sBlock+"."+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }



     /**------------------------------*
      *    read a text from a file    *
      *-------------------------------*/
     private String readFile(String sFileName)
      {
        BufferedReader fileIn=null;
        StringBuffer sb = new StringBuffer();
        String line = "" ;
        bFileOk = false ;
       
        // Open the file
        try
        {
          fileIn = new BufferedReader( new FileReader(sFileName));
        } catch (FileNotFoundException e)
        {
          System.out.println("readFile() open error: "+sFileName+" not found ");
          JOptionPane.showMessageDialog(null,"File: "+sFileName+" not found","JDAPIUtil - XML file",JOptionPane.ERROR_MESSAGE);
          return "readFile() open error: "+sFileName+" not found " ;
        }
          catch (IOException e)
        {
          System.out.println("readFile() open error: "+e);
          JOptionPane.showMessageDialog(null,"Open error: "+sFileName,"JDAPIUtil - XML file",JOptionPane.ERROR_MESSAGE);
          return "readFile() open error: "+e;
        }     
        // read the file
        try
        {
          while((line=fileIn.readLine())!= null)
          {
              sb.append(line + '\n');
          }
        }catch (Exception e)
        {
          System.out.println("readFile() reading error:"+e);
          return "readFile() reading error:"+e ;
        }
        try
        {
          fileIn.close();
        } catch (Exception e)
        {
          System.out.println("readFile() closing error:"+e);
          return "readFile() closing error:"+e ;
        }
        bFileOk = true ;
        return sb.toString(); 
      }


    /**------------------------------------------*
     *   returns an Array list of found tokens   *
     *-------------------------------------------*/
    List getXMLTagArray(String sXMLContent, String sTag)
    {
      String       sLigne      = "" ;
      String       sXml        = "" ;
      String       sTagStart   = "<" + sTag + ">" ;
      String       sTagEnd     = "</" + sTag + ">" ;
      int          iIndex      = 0;
      int          iTagDeb     = 0 ;
      int          iTagEnd     = 0 ;
      int          iTagLength  = 0 ;
      int          iLigLength  = 0 ;      
      boolean      bCont       = true ;
      
      List liste = new ArrayList(5) ; 
      
      if( sXMLContent != null && sTag != null )
      {
        sXml = sXMLContent ;
        iTagLength = sTagStart.length() ;
        while(bCont)
        {
          iIndex++ ;
          iTagEnd = sXml.indexOf(sTagEnd) ;
          iTagDeb = sXml.indexOf(sTagStart) ;        
          iLigLength = (iTagEnd - ( iTagDeb + iTagLength ) ) ;
          if(iTagDeb > -1 && iTagEnd > -1 )
          {
             //log("deb="+iTagDeb+"  end="+iTagEnd+" len="+iLigLength);
             sLigne = sXml.substring(iTagDeb+iTagLength, iTagDeb+iTagLength+iLigLength);
             liste.add((Object)sLigne.trim());
             //log("lig=["+sLigne.trim()+"]");
             int i = iTagEnd + sTagEnd.length() ;
             if(i >=sXml.length()) bCont = false ;
             else sXml = sXml.substring(i);
             //log("rest=["+sXml+"]");
          }
          else bCont = false ;
        }
      }
      return liste ;
    }
    
    /**----------------------------------*
     *    returns the content of a tag   *
     *-----------------------------------*/
    String getXMLTag(String sXMLContent, String sTag, String sNullValue)
    {
      String       sLigne      = null ;
      String       sXml        = "" ;
      String       sTagStart   = "<" + sTag + ">" ;
      String       sTagEnd     = "</" + sTag + ">" ;
      int          iTagDeb     = 0 ;
      int          iTagEnd     = 0 ;
      int          iTagLength  = 0 ;
      int          iLigLength  = 0 ;      
      
      if( sXMLContent != null && sTag != null )
      {
        sXml = sXMLContent ;
        iTagLength = sTagStart.length() ;
        iTagEnd = sXml.indexOf(sTagEnd) ;
        iTagDeb = sXml.indexOf(sTagStart) ;        
        iLigLength = (iTagEnd - ( iTagDeb + iTagLength ) ) ;
        if(iTagDeb > -1 && iTagEnd > -1 )
        {
           //log("deb="+iTagDeb+"  end="+iTagEnd+" len="+iLigLength);
           sLigne = sXml.substring(iTagDeb+iTagLength, iTagDeb+iTagLength+iLigLength);
           sLigne = sLigne.trim();
        }
      }
      if(sLigne != null && sLigne.equalsIgnoreCase("")) sLigne=null ;
      if((sLigne == null) && (sNullValue != null)) sLigne = sNullValue ;

      return sLigne ;
    }

    /**-----------------------------------------*
     *  set the new attached library case name  *
     *------------------------------------------*/
    void setLibraryCaseName(String sCase) {
       //log("-- Lib -------------------");
       //log(" setLibraryCaseName():"+sCase);
       String sName = "" ;
       int itot=0;
       // find all the module's libraries
       JdapiIterator libs = fm.getAttachedLibraries();
       libs.goStart();
       //Loop through the attached libraries
        while (libs.hasNext())   
        {  
          AttachedLibrary lib = (AttachedLibrary)libs.next();   
          itot++;  
        }        
        log(" Total libraries found:"+ itot);
        //Reloading the attached libraries  
        libs = fm.getAttachedLibraries();   
        libs.goStart();  
        String libraries[] = new String[itot+5];  
 
        int i=0;  
        // loop to fill the attached libraries names into an array of String (libraries)  
        while (libs.hasNext())  
        {  
            
          AttachedLibrary lib = (AttachedLibrary)libs.next();   
          // Identify the library name.   
          libraries[i] = lib.getName();   
          log(" -- finding lib: "+lib.getName());  
          i++;  
        }  
        
        // loop in the array to detach all libraries and reattach again in lower case  
        for (int j=0;j<itot;j++)  
        {  
          AttachedLibrary lib, lib1;  
          lib=AttachedLibrary.find(fm,libraries[j]);  
          lib.destroy();  
          log(" -- deattaching lib: "+libraries[j]);  
          if(sCase.equalsIgnoreCase("to_lower_case")) lib1 = new AttachedLibrary(fm,libraries[j].toLowerCase());   
          else if(sCase.equalsIgnoreCase("to_upper_case")) lib1 = new AttachedLibrary(fm,libraries[j].toUpperCase());   
          log(" -- reattaching lib: "+libraries[j].toLowerCase());  
        }           
    }

    // -------------------
    //  scan for triggers
    // -------------------
    boolean scanTriggers(String sSource)
    {
        List l = getXMLTagArray(sSource, "trigger") ;
        boolean b=true ;
        if(l.size()>0)
        {
          for(int j=0; j<l.size(); j++)
          {
            String st = (String)l.get(j) ;
            //log("TRG-->["+st2+"]");
            String sAction = getXMLTag(st, "action", null) ;
            String sName   = getXMLTag(st, "name", null) ;
            String sRepl   = getXMLTag(st, "replace", null) ;
            String sText   = getXMLTag(st, "text", null) ;
            String sPos    = getXMLTag(st, "position", "start") ;
            String sErr    = getXMLTag(st, "on_error", "continue") ;
            //log("-->"+sAction+"-"+sName+"- replace["+sRepl+"] by ["+sText+"]");
            b = doAction("TRIGGER", sAction, sName, sRepl, sText, sPos, sErr) ;
          }
        }
        return b ;
    }
    
    // ------------------------
    //  scan for program units
    // ------------------------
    boolean scanProgramUnits(String sSource)
    {
        List l = getXMLTagArray(sSource, "program_unit") ;
        boolean b=true ;
        if(l.size()>0)
        {
          for(int j=0; j<l.size(); j++)
          {
            String st = (String)l.get(j) ;
            //log("PU-->["+st2+"]");
            String sAction = getXMLTag(st, "action", null) ;
            String sName   = getXMLTag(st, "name", null) ;
            String sRepl   = getXMLTag(st, "replace", null) ;
            String sText   = getXMLTag(st, "text", null) ;
            String sPos    = getXMLTag(st, "position", "start") ;
            String sErr    = getXMLTag(st, "on_error", "continue") ;
            //log("-->"+sAction+"-"+sName+"-"+sText);
            b = doAction("PROCEDURE", sAction, sName, sRepl, sText, sPos, sErr) ;
          }
        }            
        return b ;
    }
    
    // ------------------------------
    //  scan for attached libraries
    // ------------------------------
    boolean scanLibraries(String sSource)
    {
        List l = getXMLTagArray(sSource, "library") ;
        boolean b=true ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
              String st = (String)l.get(j) ;
              String sAction = getXMLTag(st, "action", null) ;
              String sName   = getXMLTag(st, "name", null) ;
              String sErr    = getXMLTag(st, "on_error", "continue") ;
              //log("-->"+sAction+"-"+sName+"-"+sText);
              b = doAction("LIBRARY", sAction, sName, "", "", "", sErr) ;
            }
        }            
        return b ;
    }
    
    // ------------------------------
    //  scan for subclassed objects
    // ------------------------------
    boolean scanSubclassedObjects(String sSource)
    {
        List l = getXMLTagArray(sSource, "subclass_olb") ;
        boolean b=true ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                String sName   = getXMLTag(st, "lib_name", null) ;
                String sTab    = getXMLTag(st, "tab_name", null) ;
                String sObject = getXMLTag(st, "object_num", null) ;              
                String sLabel  = getXMLTag(st, "target_name", null) ;
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("OLB", "", sName, sTab, sObject, sLabel, sErr) ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //   scan for blocks
    // --------------------
    boolean scanBlocks(String sSource,String sBlockName)
    {
        List l = getXMLTagArray(sSource, "block") ;
        boolean b=true ;
        String sName = sBlockName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("BLOCK", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }

    
    // --------------------
    //  scan for canvases
    // --------------------
    boolean scanCanvases(String sSource,String sCanvasName)
    {
        List l = getXMLTagArray(sSource, "canvas") ;
        boolean b=true ;
        String sName = sCanvasName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("scanCanvases-->"+sName+"-"+sProp+"-"+sValue);
                b = doAction("CANVAS", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //  scan for windows
    // --------------------
    boolean scanWindows(String sSource,String sWindowName)
    {
        List l = getXMLTagArray(sSource, "window") ;
        boolean b=true ;
        String sName = sWindowName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("WINDOW", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //  scan for alerts
    // --------------------
    boolean scanAlerts(String sSource,String sAlertName)
    {
        List l = getXMLTagArray(sSource, "alert") ;
        boolean b=true ;
        String sName = sAlertName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("ALERT", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }

    
    // ----------------------
    //  scan for parameters
    // ----------------------
    boolean scanParameters(String sSource,String sParamName)
    {
        List l = getXMLTagArray(sSource, "parameter") ;
        boolean b=true ;
        String sName = "" ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                //if(sName == null) sName   = getXMLTag(st, "name", null) ;
                sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "action", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                b = doAction("PARAMETER", sProp, sName.toUpperCase(), "", sValue, sErr, "") ;
            }
        }            
        return b ;
    }

    
    // -----------------
    //  scan for items
    // -----------------
    boolean scanItems(String sSource,String sBlockName, String sItemName)
    {
        List l = getXMLTagArray(sSource, "item") ;
        boolean b=true ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                String sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("WINDOW", "update", sName, "", sProp, sValue, sErr) ;
            }
        }            
        return b ;
    }

   /**------------------------*
    *   open a Forms module   *
    *-------------------------*/
   boolean openModule(String sModuleName)
   {
        String s = "-- proceeding module: " + sModuleName + " --" ;
        log(" ");
        log(pad('-',s.length()));
        log(s);
        log(pad('-',s.length()));

        f = new File(sModuleName);
        // path
        sPath = f.getParent()+f.separator;
        // file name
        sFile = f.getName();
        // file extension
        ie = sFile.indexOf(".") ;
        sExt = sFile.substring(ie+1);
        sFile = sFile.substring(0,ie);
        // extension must be .fmb
        if( ! sExt.equalsIgnoreCase("FMB"))
        {
          System.setErr(System.out);
          printError("Module : "+sModuleName+". File extension must be .FMB");
          return false ;
        }

        try{
          mod = JdapiModule.openModule(sModuleName);             
          fm = (FormModule)mod.getModule();
        }
        catch(Exception ex)
        {
          System.setErr(System.out);
          printError("Module : "+sModuleName+". Unable to open. Error details: " + ex.toString());
          return false ;
        }
        return true ;
    }
    
    /**---------------------*
     *   compile a module   *
     *----------------------*/
    boolean compileModule(String sFile)
    {
        log("-- Compilation of module:"+sFile);
        try {
          Jdapi.connectToDatabase(sConnect); 
        }
        catch (JdapiException e) 
        {
          System.setErr(System.out);
          printError("Database connection : Unable to connect to the specified database. The form has been modified but not compiled. Please compile it manually. Error details: " + e.toString());
          return false ;
        }
        // Compile the form to create a fmx. 
        if (Jdapi.isConnectedToDatabase() ) {
          try{
            fm.compile(); 
            log("   module: "+sFile+ " compiled.");
          }
          catch (JdapiException e) 
          {
            System.setErr(System.out);
            printError("Compilation : Unable to compile The form. Error details: " + e.toString());
            return false ;
          }
        }
        return true ;
    }

    /**-----------------------------------*
     *  replaceAll() kind 1.3 equivalent  *
     *------------------------------------*/
    String _replaceAll(String source, String pattern, String replace)
    {
        if (source!=null)
        {
        final int len = pattern.length();
        StringBuffer sb = new StringBuffer();
        int found = -1;
        int start = 0; 
        while( (found = source.indexOf(pattern, start) ) != -1) {
            sb.append(source.substring(start, found));
            sb.append(replace);
            start = found + len;
        }

        sb.append(source.substring(start));

        return sb.toString();
        }
        else return "";
    }


    /*-------------------------------------------*
     *   Object to handle canvases information   *
     *-------------------------------------------*/
    class cCanvas extends Object 
    {
        String      sName      = "" ;
        String      sBeanName  = "" ;
        int         iNbre = 0 ;

        cCanvas(String p_sCanvasName, String p_sBeanName) {
            sName      = p_sCanvasName ;
            sBeanName  = p_sBeanName ;
        }        
        cCanvas(String p_sCanvasName, int p_iNbre) {
            sName      = p_sCanvasName ;
            iNbre      = p_iNbre ;
        }        
        void add() { iNbre++ ; }
    }    


    /*---------------------------------------------------*
     *   Object to handle blocks' canvases information   *
     *---------------------------------------------------*/
    class cBlock extends Object 
    {
        String      sName      = "" ;
        String      sCanvasName  = "" ;
        int         iNbre = 0 ;

        cBlock(String p_sBlockName, String p_sCanvasName) {
            sName        = p_sBlockName ;
            sCanvasName  = p_sCanvasName ;
        }        
        void add() { iNbre++ ; }
    }    

    
    public static boolean getBooleanValue(String sValue)
    {
      if(sValue.equalsIgnoreCase("true"))  return true ;
      else return false ;
    }
    
    /**----------------------------*
     *  Ouput messages on console  *
     *-----------------------------*/
    private void log( String msg)
    {
       if(bLog) System.out.println(msg) ;
       if(bLogOpen) writeLogFile(msg);
    }
    
    private void log3( String msg)
    {
       if(bShowListComps) System.out.println(msg) ;
       if(bLogOpen && bLogDetail) writeLogFile(msg);
    }     
    
    private void log2( String msg)
    {
       if (!bShowListComps){
         if(bLog) System.out.println(msg) ;
         if(bLogOpen && bLogDetail) writeLogFile(msg);
       }
    } 
    /**------------------*
     *   open log file   *
     *-------------------*/
    boolean openLogFile(String sFileName)
    {
      // open the file
      log("Opening log file: " + sFileName ) ;
      try
      {
        fileOut = new PrintWriter(new FileWriter(sFileName));
        bLogOpen = true ;
      } catch (Exception e)
      {
        printError("openLogFile opening error:"+e); 
        bLogOpen = false ;
        return false ;
      }  
      return true ;
    }
    
    /**-------------------*
     *   close log file   *
     *--------------------*/
    void closeLogFile()
    {
       try
       {
         if(fileOut != null) fileOut.close();
       } catch (Exception e)
       {
         printError("closeLogFile closing error:"+e);
       }       
       finally{bLogOpen = false ;}
    }
    
  /**----------------------*
   *   Write to log file   *
   *-----------------------*/
  private void writeLogFile(String sLine)
  {
      // write to the file
      try
        {
          fileOut.println( sLine ) ;
        } catch (Exception e)
        {
          printError(" writeLogFile writing error:"+e); 
        }
  }
        
   private void printGraphicProperties(Graphics g)
   {
     System.out.println("   .Owner             "+g.getOwner());  
     System.out.println("   .Name              "+g.getName());       
     System.out.println("   .Title             "+g.getFrameTitle());  
     System.out.println("   .Frame align       "+g.getFrameAlign());  
     System.out.println("   .Frame Title align "+g.getFrameTitleAlign());  
     System.out.println("   .Font name         "+g.getFrameTitleFontName());  
     System.out.println("   .Font size         "+g.getFrameTitleFontSize());  
     System.out.println("   .Height            "+g.getHeight());  
     System.out.println("   .Width             "+g.getWidth());  
     System.out.println("   .X pos             "+g.getXPosition());  
     System.out.println("   .Y pos             "+g.getYPosition());  
     System.out.println("   .ScrollBar         "+g.isShowScrollbar());       
     System.out.println("   .Parent name       "+g.getParentName());  
     System.out.println("   .Records displayed "+g.getRecordsDisplayCount());       
     System.out.println("   .Source            "+g.getSource());  
     System.out.println("   .Data Block        "+g.getLayoutDataBlockName());  
   }
   
   
   /**------------------------------------------------*
    *  Simply prints an error message to the console  *
    *-------------------------------------------------*/
   private void printError(String errorString)
   {
     System.out.println("  *** " + sCurrentModule + " ERROR *** " + errorString);
   }

    private String pad(char c1, int i1)
    {
      StringBuffer sb = new StringBuffer();
      for(int i=0 ; i < i1; i++) sb.append(c1);
      return sb.toString() ;
    }
    private String rpad(String s1, char c1, int i1)
    {
      StringBuffer sb = new StringBuffer(s1);
      for(int i=0 ; i < i1; i++) sb.append(c1);
      return sb.toString() ;
    }    
    private String lpad(String s1, char c1, int i1)
    {
      StringBuffer sb = new StringBuffer();
      for(int i=0 ; i < i1; i++) sb.append(c1);
      sb.append(s1);
      return sb.toString() ;
    }        
    
} 
 

